#include <string.h>
#include <libxml/parser.h>

#include "main.h"
#include "log.h"
#include "pack.h"
#include "ship.h"

#define MAX_PATH_NAME 30 // Maximum size of the path.

#define XML_NODE_START  1
#define XML_NODE_TEXT   3

#define XML_ID    "Ships" // XML section identifier.
#define XML_SHIP  "ship"

#define SHIP_DATA "../dat/ship.xml"
#define SHIP_GFX  "../gfx/ship/"

static Ship* ship_stack = NULL;
static int ships = 0;

// Get a ship based on it's name.
Ship* get_ship(const char* name) {
  Ship* tmp = ship_stack;
  int i;
  for(i = 0; i < ships; i++)
    if(strcmp((tmp+i)->name, name)==0) break;
  
  if(i == ships) // Ship doesn't exist, game will probably crash now.
    WARN("Ship %s does not exist", name);

  return tmp+i;
}

Ship* ship_parse(xmlNodePtr parent) {
  xmlNodePtr cur, node;
  Ship* tmp = CALLOC_L(Ship);

  char str[MAX_PATH_NAME] = "\0";

  tmp->name = (char*)xmlGetProp(parent, (xmlChar*)"name");

  node = parent->xmlChildrenNode;

  while((node = node->next)) { // Load all the data.
    if(strcmp((char*)node->name, "GFX")==0) {
      cur = node->children;
      if(strcmp((char*)cur->name, "text")==0) {
        snprintf(str, strlen((char*)cur->content)+sizeof(SHIP_GFX),
              SHIP_GFX"%s", (char*)cur->content);
        tmp->gfx_ship = gl_newSprite(str, 6, 6);
      }
    }
    else if(strcmp((char*)node->name, "class")==0) {
      cur = node->children;
      if(strcmp((char*)cur->name, "text")==0)
        tmp->class = atoi((char*)cur->content);
    }
    else if(strcmp((char*)node->name, "movement")==0) {
      cur = node->children;
      while((cur = cur->next)) {
        if(strcmp((char*)cur->name, "thrust")==0)
          tmp->thrust = atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "turn")==0)
          tmp->turn = atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "speed")==0)
          tmp->speed = atoi((char*)cur->children->content);
      }
    }
    else if(strcmp((char*)node->name, "health")==0) {
      cur = node->children;
      while((cur = cur->next)) {
        if(strcmp((char*)cur->name, "armor")==0)
          tmp->armor = (double)atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "shield")==0)
          tmp->shield = (double)atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "energy")==0)
          tmp->energy = (double)atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "armor_regen")==0)
          tmp->armor_regen = (double)(atoi((char*)cur->children->content))/60.0;
        else if(strcmp((char*)cur->name, "shield_regen")==0)
          tmp->shield_regen = (double)(atoi((char*)cur->children->content))/60.0;
        else if(strcmp((char*)cur->name, "energy_regen")==0)
          tmp->energy_regen = (double)(atoi((char*)cur->children->content))/60.0;
      }
    }
    else if(strcmp((char*)node->name, "characteristics")==0) {
      cur = node->children;
      while((cur = cur->next)) {
        if(strcmp((char*)cur->name, "crew")==0)
          tmp->crew = atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "mass")==0)
          tmp->mass = (double)atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "cap_weapon")==0)
          tmp->cap_weapon = atoi((char*)cur->children->content);
        else if(strcmp((char*)cur->name, "cap_cargo")==0)
          tmp->cap_cargo = atoi((char*)cur->children->content);
      }
    }
  }
  tmp->thrust *= tmp->mass; // Helps keep number sane.

#define MELEMENT(o,s) if(o == 0) WARN("Ship '%s' missing '"s"' element", tmp->name)
  if(tmp->name == NULL) WARN("Ship '%s' missing 'name' tag", tmp->name);
  if(tmp->gfx_ship == NULL) WARN("Ship '%s' missing 'GFX' element", tmp->name);
  MELEMENT(tmp->thrust,       "thrust");
  MELEMENT(tmp->turn,         "turn");
  MELEMENT(tmp->speed,        "speed");
  MELEMENT(tmp->crew,         "crew");
  MELEMENT(tmp->mass,         "mass");
  MELEMENT(tmp->armor,        "armor");
  MELEMENT(tmp->armor_regen,  "armor_regen");
  MELEMENT(tmp->shield,       "shield");
  MELEMENT(tmp->shield_regen, "shield_regen");
  MELEMENT(tmp->energy,       "energy");
  MELEMENT(tmp->energy_regen, "energt_regen");
  MELEMENT(tmp->cap_cargo,    "cap_cargo");
  MELEMENT(tmp->cap_weapon,   "cap_weapon");
#undef MELEMENT

  DEBUG("Loaded ship '%s'", tmp->name);
  return tmp;
}

int ships_load(void) {
  uint32_t bufsize;
  char* buf = pack_readfile(DATA, SHIP_DATA, &bufsize);

  xmlNodePtr node;
  xmlDocPtr doc = xmlParseMemory(buf, bufsize);
  
  Ship* tmp = NULL;

  node = doc->xmlChildrenNode; // Ships node.
  if(strcmp((char*)node->name, XML_ID)) {
    ERR("Malformed "SHIP_DATA" file: missing root element '"XML_ID"'");
    return -1;
  }

  node = node->xmlChildrenNode; // First ship node.
  if(node == NULL) {
    ERR("Malformed "SHIP_DATA" file: Does not contain elements");
    return -1;
  }
  
  do {
    if(node->type == XML_NODE_START && strcmp((char*)node->name, XML_SHIP)==0) {
      tmp = ship_parse(node);
      ship_stack = realloc(ship_stack, sizeof(Ship)*(++ships));
      memcpy(ship_stack+ships-1, tmp, sizeof(Ship));
      free(tmp);
    }
  } while((node = node->next));

  xmlFreeDoc(doc);
  free(buf);
  xmlCleanupParser();

  return 0;
}

void ships_free(void) {
  int i;
  for(i = 0; i < ships; i++) {
    if((ship_stack+i)->name)
      free((ship_stack+i)->name);
    gl_freeTexture((ship_stack+i)->gfx_ship);
  }
  free(ship_stack);
  ship_stack = NULL;
}

